#import "gdb.h"

#import "LazyRegionManager.h"
#import "terminal.h"
#import <mach/cthreads.h>
#import <mach/notify.h>
#import <mach/exception.h>

#import <objc/objc-runtime.h>

#import <sys/ptrace.h>
#import <sys/signal.h>
#import <sys/wait.h>

extern struct symbol *step_start_function;
extern int stop_print_frame;
extern int stop_soon_quietly;
extern int breakpoints_inserted;
extern int breakpoints_failed;

static cthread_t signalThread = 0;
static port_t signalPort = PORT_NULL;
static mutex_t childrenMutex = 0;
static condition_t childrenCondition;
static BOOL childrenExist;
static port_t gdb_notify_port;
static port_set_name_t gdb_ports;
static port_t inferior_exception_port, inferior_old_exception_port;
static int signalSuspendCount = 0;
static int lazyRead = 0;
static BOOL handle_exception;
static BOOL areStepping = 0;
const static char breakpoint_insn[] = BREAKPOINT;
static char shadow[sizeof(breakpoint_insn)];

#ifdef IMPRECISE_EXECEPTIONS			        
static CORE_ADDR breakPC = 0;
#endif /* IMPRECISE_EXECEPTIONS */

thread_t catch_thread;
int catch_exception;
int catch_code;
int catch_subcode;
task_t catch_task;
struct breakpoint *catch_data_break = NULL;
RegionManager *regionManager;
task_t inferior_task;
thread_t current_thread = 0, step_thread = 0, last_current_thread = 0;
int kernel_attached = NO;
boolean_t stopped_in_ptrace;
int read_memory_out_of_bounds_ok = 0;

void create_inferior_for_task(task_t task, BOOL isLazy);
void create_inferior_hook(char *args, int pid);
void NeXT_mourn_inferior ();

/* This function simply calls ptrace with the given arguments.  
   It exists so that all calls to ptrace are isolated in this 
   machine-dependent file. */
int
call_ptrace (int request, int pid, int arg3, int arg4)
{
  return ptrace (request, pid, arg3, arg4);
}

CORE_ADDR getImp(CORE_ADDR pc)
{
    unsigned object, sel; 
    if (getFirstTwoIntArgs(&object, &sel)) {
	CORE_ADDR imp = find_implementation(object, sel);
	return imp;
    } else
	return 0;
}

CORE_ADDR getImpForMsgSuper(CORE_ADDR pc)
{
    unsigned sel, super; 
    if (getFirstTwoIntArgs(&super, &sel)) {
	struct objc_super *mySuper = WARP((struct objc_super *)super);
	Class class = mySuper ? mySuper->class : NULL;
	CORE_ADDR imp = class ? find_implementation_from_class(class, sel)
	                      : 0;
	return imp;
    } else
	return 0;
}

typedef struct _Ignore {
    char *name;
    CORE_ADDR (*stop_at)(CORE_ADDR);
    CORE_ADDR begin;
    CORE_ADDR end;
} Ignore;

static Ignore ignored[] = {
   {"objc_msgSend",                &getImp,            0, 0},
   {"objc_msgSendSuper",           &getImpForMsgSuper, 0, 0},
   {"objc_getClass",               NULL,               0, 0},
   {"objc_getMetaClass",           NULL,               0, 0}
};
static int numIgnored = 0;
#define NUMIGNORED (sizeof(ignored) / sizeof(*ignored))
static int foundIndeces = NO;

static inline BOOL
ISLIBPREFIX(char *name)
{
    return (((*name == '.') || (*name == '_'))
            && (*(name + 1) == 'l')
	    && (*(name + 2) == 'i')
	    && (*(name + 3) == 'b'));
}

int
pc_off_limits (pc, name, new_pc)
CORE_ADDR pc;
char *name;
CORE_ADDR *new_pc;
{
    int i;
    CORE_ADDR new_stop = 0, new_new_stop;
    int off_limits = 0;
    struct minimal_symbol *nextFunc;
    char *underScore;
    
    if (!isShlibJump(pc)) {
        off_limits = YES;
	new_stop = getBranch(pc);
    } else if (name
               && (ISLIBPREFIX(name))
	       && ((underScore = strchr(name, '_'))
	           || (underScore = strchr(name, '.')))
	       && (nextFunc = lookup_minimal_symbol(underScore + 1, NULL))) {
	off_limits = YES;
	new_stop = nextFunc->address;
    } else {
        if (!foundIndeces) {
	    for (i = 0; i < NUMIGNORED; i++) {
//		ignored[i].name = stringSave(ignored[i].name);
		nextFunc = lookup_minimal_symbol(ignored[i].name, NULL);
		if (nextFunc) {
		    ignored[i].begin = nextFunc->address;
		    for (;
		         nextFunc->address
			 == (nextFunc + 1)->address;
			 nextFunc++);
		    ignored[numIgnored].end = (nextFunc + 1)->address;
		    numIgnored = i + 1;
		}
	    }
	    foundIndeces = YES;
	}
	for (i = 0; i < numIgnored; i++) {
	    if ((name
	         && (*ignored[i].name == *name)
		 && (strcmp(ignored[i].name + 1, name + 1) == 0))
	        || (!name
		    && ((ignored[i].begin <= pc) && (pc < ignored[i].end)))) {
	        off_limits = YES;
		if (ignored[i].stop_at)
		    new_stop = (*ignored[i].stop_at)(pc);
	    }
	}
    }
    if (new_stop) {
        if (pc_off_limits(new_stop, NULL, &new_new_stop))
	    *new_pc = new_new_stop;
	else {
	    if (find_pc_function(new_stop))
		*new_pc = new_stop;
	    else
		*new_pc = 0;
	}
    } else 
	*new_pc = 0;
    return off_limits;
}

int process_notify_msg(notification_t *msg)
{
    if ((msg->notify_header.msg_id == NOTIFY_PORT_DELETED)
	&& (msg->notify_port == inferior_task))
	return 1;
    else
	return 0;
}

kern_return_t
catch_exception_raise(port_t port, thread_t thread, task_t task,
                          int exception, int code, int subcode)
{
    port_t replyPort;
//    mach_NeXT_exception("catch_exception_raise", exception, code, subcode);
    catch_task = task;
    catch_thread = thread;
    catch_exception = exception;
    catch_code = code;
    catch_subcode = subcode;
    if (task == inferior_task) {
	kern_return_t ret = task_suspend(task);
	handle_exception = YES;
	mach_call(ret, "task_suspend", "catch_exception_raise");
        return ret;
    } else {
	handle_exception = NO;
	return KERN_FAILURE;
    }
}

struct msg {
  msg_header_t    header;
  int             data[20];
};

typedef struct _SignalMessage {
    msg_header_t header;
    msg_type_t   type;
    int          pid;
    int		 status;
} SignalMsg;

#define SIGNAL_MSG 0
#define STOP_MSG 1

void putExceptionPortBack(int dummy)
{
    port_set_add(task_self(), gdb_ports, inferior_exception_port);
}

void drain_exceptions()
{
    struct msg inMsg, outMsg;
    int noMore, ret;
    struct cleanup *oldChain;
    
    mach_call_no_inferior(port_set_remove(task_self(), inferior_exception_port),
                          "port_set_remove", "drain_exceptions");
    oldChain = make_cleanup(putExceptionPortBack, 0);
    for (noMore = NO; !noMore;) {
	inMsg.header.msg_local_port = inferior_exception_port;
	inMsg.header.msg_size = sizeof(inMsg);
	ret = msg_receive(&inMsg.header, RCV_TIMEOUT, 0);
	if (ret == KERN_SUCCESS) {
	    debugPrintf("Before exc_server, incoming remote port is %x.\n",
	                 inMsg.header.msg_remote_port);
            ret = exc_server(&inMsg, &outMsg);
	    debugPrintf("exc_server returned %d.\n", ret);
	    debugPrintf("After exc_server, outgoing remote port is %x.\n",
	                 outMsg.header.msg_remote_port);
	    if (ret && handle_exception) {
		msg_send(&outMsg.header, MSG_OPTION_NONE, 0);
		task_resume(inferior_task);
		if (catch_exception != EXC_BREAKPOINT) {
                    char exception_string[100];
                    strcpy(exception_string,
                            mach_NeXT_exception_string(catch_exception,
                                    catch_code,
                                    catch_subcode));
                    terminal_ours_for_output ();
                    printf("Program generated(%d): %s.\n",
                           catch_exception,
                           exception_string);
                }
	    } else {
		exception_raise(inferior_old_exception_port, 
		                thread_reply(),
	                        catch_thread, catch_task, catch_exception, 
				catch_code, catch_subcode);
	    }
	} else 
	    noMore = YES;
    }
    mach_call_no_inferior(
      port_set_add(task_self(), gdb_ports, inferior_exception_port),
      "port_set_add", "create_inferior_hook");
    discard_cleanups(oldChain);
}

#if 0
void
child_create_inferior(char *exec_file, char *allargs, char **env)
{
  int pid;
  char *shell_command;
  extern int sys_nerr;
  extern char *sys_errlist[];
  extern struct target_ops child_ops;
  int len;
  int pending_execs;
  int debug_setpgrp;
  int fildes[2];
  char dummy = 0;
  
  len = 5 + strlen (exec_file) + 1 + strlen (allargs) + 1 + /*slop*/ 10;
  shell_command = (char *) alloca (len);
  sprintf(shell_command, "exec %s %s", exec_file, allargs);
  
  pipe(fildes);

  pid = vfork ();

  if (pid < 0)
    perror_with_name ("vfork");

  if (pid == 0)
    {
#ifdef TIOCGPGRP
      /* Run inferior in a separate process group.  */
      debug_setpgrp = setpgrp (getpid (), getpid ());
      if (0 != debug_setpgrp)
	 perror("setpgrp failed in child");
#endif /* TIOCGPGRP */

      /* Tell the terminal handling subsystem what tty we plan to run on;
	 it will now switch to that one if non-null.  */

      new_tty ();

      /* Changing the signal handlers for the inferior after
	 a vfork can also change them for the superior, so we don't mess
	 with signals here.  See comments in
	 initialize_signals for how we get the right signal handlers
	 for the inferior.  */

      read(fildes[0], &dummy, 1);
      close(fildes[0]);
      call_ptrace (0, 0, 0, 0);		/* "Trace me, Dr. Memory!" */
      execle ("/bin/sh", "/bin/sh", "-c", shell_command, (char *)0, env);

      fprintf (stderr, "Cannot exec %s: %s.\n", "/bin/sh",
	       errno < sys_nerr ? sys_errlist[errno] : "unknown error");
      fflush (stderr);
      _exit (0177);
    }

  /* Now that we have a child process, make it our target.  */
  push_target (&child_ops);

  create_inferior_hook(NULL, pid);

/* The process was started by the fork that created it,
   but it will have stopped one instruction after execing the shell.
   Here we must get it up to actual execution of the real program.  */

  inferior_pid = pid;		/* Needed for wait_for_inferior stuff below */

  clear_proceed_status ();

  /* We will get a trace trap after one instruction.
     Continue it automatically.  Eventually (after shell does an exec)
     it will get another trace trap.  Then insert breakpoints and continue.  */

  pending_execs = 2;

  init_wait_for_inferior ();

  /* Set up the "saved terminal modes" of the inferior
     based on what modes we are starting it with.  */
  target_terminal_init ();

  /* Install inferior's terminal modes.  */
  target_terminal_inferior ();
  
  write(fildes[1], &dummy, 1);
  close(fildes[1]);
  task_resume(inferior_task);

  while (1)
    {
      stop_soon_quietly = 1;	/* Make wait_for_inferior be quiet */
      wait_for_inferior ();
      if (stop_signal != SIGTRAP)
	{
	  /* Let shell child handle its own signals in its own way */
	  /* FIXME, what if child has exit()ed?  Must exit loop somehow */
	  resume (0, stop_signal);
	}
      else
	{
	  /* We handle SIGTRAP, however; it means child did an exec.  */
	  if (0 == --pending_execs)
	    break;
	  resume (0, 0);		/* Just make it go on */
	}
    }
  stop_soon_quietly = 0;
  
  initLoaded();

  /* Should this perhaps just be a "proceed" call?  FIXME */
  insert_step_breakpoint ();
  insertLoadBreakpoint();
  breakpoints_failed = insert_breakpoints ();
  if (!breakpoints_failed)
    {
      breakpoints_inserted = 1;
      target_terminal_inferior();
      /* Start the child program going on its first instruction, single-
	 stepping if we need to.  */
      resume (bpstat_should_step (), 0);
      wait_for_inferior ();
      normal_stop ();
    }
}

void
NeXT_create_inferior(char *execFile, char *allArgs, char**env)
{
    child_create_inferior(execFile, allArgs, env);
}

#endif 0

int mach_oh_wait(int *status)
{
    int ret, gotMsg = NO, isGone = NO;
    struct msg inMsg, outMsg;
    union wait *w = (union wait *)status;
    SignalMsg *signalMsg;
    struct breakpoint *db = NULL;
    BOOL stepping = areStepping;
    static CORE_ADDR retPC = 0;
    extern BOOL safetiesInserted;

    stopped_in_ptrace = NO;
    while (!gotMsg) {
	inMsg.header.msg_local_port = gdb_ports;
	inMsg.header.msg_size = sizeof(inMsg);
	ret = msg_receive(&inMsg.header, 0, 0);
        if (ret == KERN_SUCCESS) {
	    if (inMsg.header.msg_local_port == gdb_notify_port) {
		if (process_notify_msg((notification_t *)&inMsg)) {
		    if (attach_flag) {
			gotMsg = YES;
			isGone = PROCESS_DISAPPEARED;
		    }
		}
	    } else if (inMsg.header.msg_local_port
	               == inferior_exception_port) {
		debugPrintf("Before exc_server, incoming remote port is %x.\n",
			     inMsg.header.msg_remote_port);
		ret = exc_server(&inMsg, &outMsg);
		debugPrintf("exc_server returned %d.\n", ret);
		debugPrintf("After exc_server, outgoing remote port is %x.\n",
				outMsg.header.msg_remote_port);
		if (ret && handle_exception) {
		    mach_call_no_inferior(
			msg_send(&outMsg.header, MSG_OPTION_NONE, 0),
			"msg_send", "mach_oh_wait");
		    prepare_threads_after_stop();
		    if (catch_exception == EXC_BAD_ACCESS) {
			db = is_data_break(catch_subcode);
#ifdef IMPRECISE_EXECEPTIONS			    
			if (db)
			    breakPC = read_pc();
#endif /* IMPRECISE_EXECEPTIONS	*/
			if (db == (struct breakpoint *)0xFFFFFFFF) {
			    remove_breakpoints ();
			    breakpoints_inserted = 0;
			    registers_changed();
			    child_resume(1, 0);
			} else if (db != 0) {
			    catch_data_break = db;
			    gotMsg = YES;
			} else
			    gotMsg = YES;
		    } else if (catch_exception == EXC_BREAKPOINT) {
#ifdef IMPRECISE_EXECEPTIONS
			if (breakPC) {			
			    [regionManager putDataAt: (void *)breakPC
						 for: sizeof(breakpoint_insn)
						from: (void *)shadow];
			    write_pc(read_pc() - DECR_PC_AFTER_BREAK);
			    breakPC = 0;
			}
#endif /* IMPRECISE_EXECEPTIONS */			    
			if (db) {
			    db = NULL;
			    registers_changed();
			    if (stepping)
				child_resume(1, 0);
			    else {
				insert_breakpoints ();
				breakpoints_inserted = 1;
				child_resume(0, 0);
			    }
			} else if (safetiesInserted
			           && isSafetyBreakpoint(read_pc() - DECR_PC_AFTER_BREAK)) {
			    retPC = INITIAL_RET_PC();
			    removeSafetyBreakpoints();
			    [regionManager flushMarkedPages];
			    remove_breakpoints ();
			    breakpoints_inserted = 0;
                            [regionManager getDataAt: (void *)retPC
						 for: sizeof(breakpoint_insn)
						into: (void *)shadow];
			    [regionManager putDataAt: (void *)retPC
						 for: sizeof(breakpoint_insn)
						from: (void *)breakpoint_insn];
			    write_pc(read_pc() - DECR_PC_AFTER_BREAK);
			    registers_changed();
			    child_resume(0, 0);
			} else if (retPC 
			          && (read_pc() - DECR_PC_AFTER_BREAK
				      == retPC)) {
			    [regionManager putDataAt: (void *)retPC
						 for: sizeof(breakpoint_insn)
						from: (void *)shadow];
			    retPC = 0;
			    insert_breakpoints ();
			    breakpoints_inserted = 1;
			    write_pc(read_pc() - DECR_PC_AFTER_BREAK);
			    registers_changed();
			    child_resume(0, 0);
			} else
			    gotMsg = YES;
		    } else
			gotMsg = YES;
			
		    if (gotMsg && (catch_thread != current_thread)) {
			flush_cached_frames();
			last_current_thread = current_thread;
			current_thread = catch_thread;
		    }
		} else {
		    printf("The old exception port is %p\n", inferior_old_exception_port);
		    exception_raise(inferior_old_exception_port, 
		                    thread_reply(),
	                            catch_thread, catch_task, catch_exception, 
				    catch_code, catch_subcode);
		}
	    } else if (inMsg.header.msg_local_port == signalPort) {
		signalMsg = (SignalMsg *)&inMsg;
		if (signalMsg->pid == inferior_pid) {
		    if (signalMsg->header.msg_id == SIGNAL_MSG) {
			*status = signalMsg->status;
//	                printf("Heard about signal %d for pid %d.\n", w->w_stopsig, inferior_pid);
			gotMsg = YES;
			if (WIFEXITED(*w))
			    isGone = PROCESS_EXITED;
			else if (!WIFSTOPPED(*w))
			    isGone = PROCESS_TERMINATED;
//			else
//			    prepare_threads_after_stop();
			stopped_in_ptrace = YES;
		    } else if (signalMsg->header.msg_id == STOP_MSG){
			gotMsg = YES;
			mach_call(task_suspend(inferior_task),
				  "task_suspend", "mach_oh_wait");
			drain_exceptions();
		        prepare_threads_after_stop();
			catch_exception = -1;
		    }
		}
	    }
	}
    }
    return (int)isGone;
}

void buildSignalMsg(SignalMsg *signalMsg, int msg_id)
{
    signalMsg->header.msg_simple = TRUE;
    signalMsg->header.msg_size = sizeof(SignalMsg);
    signalMsg->header.msg_type = MSG_TYPE_NORMAL;
    signalMsg->header.msg_local_port = PORT_NULL;
    signalMsg->header.msg_remote_port = signalPort;
    signalMsg->header.msg_id = msg_id;

    signalMsg->type.msg_type_name = MSG_TYPE_INTEGER_32;
    signalMsg->type.msg_type_size = 32;
    signalMsg->type.msg_type_number = 2;
    signalMsg->type.msg_type_inline = TRUE;
    signalMsg->type.msg_type_longform = FALSE;
    signalMsg->type.msg_type_deallocate = FALSE;
}

void
waitForSignal(void *dummy)
{
    union wait w;
    int waitPid;
    SignalMsg signalMsg;
    buildSignalMsg(&signalMsg, SIGNAL_MSG);
    while (1) {
//	printf("Waiting for signals.\n");
	waitPid = wait(&w);
	if (waitPid == -1) {
//	    printf("There are no children to wait on.\n");
	    mutex_lock(childrenMutex);
	    childrenExist = NO;
	    while (!childrenExist)
		condition_wait(childrenCondition, childrenMutex);
	    mutex_unlock(childrenMutex);
	} else {
//	    printf("Got signal %d for pid %d.\n", w.w_stopsig, waitPid);
	    signalMsg.pid = waitPid;
	    signalMsg.status = w.w_status;
	    msg_send(&signalMsg.header, 0, 0);
	}
    }
}

void
inferiorSigInt()
{
    SignalMsg signalMsg;
    
    buildSignalMsg(&signalMsg, STOP_MSG);
    signalMsg.pid = inferior_pid;
    msg_send(&signalMsg.header, 0, 0);
}

void
kill_inferior_fast ()
{
    kern_return_t ret;
    int w;
    if (inferior_pid == 0)
	return;
    else if (inferior_pid == -1) {
	unloadStatesFrom(0);
	if (mach_call_no_error(task_terminate(inferior_task),
	                       "task_terminate", "kill_inferior_fast"))
	    return;
    } else {
	unloadStatesFrom(0);
	voyeurResume();
#ifdef i386
 	//
 	// BKM - XXX 4/20/92
 	// This is needed to prevent a race condition between the child's
 	// death and possible arrival of a breakpoint exception (which would
 	// make our exception handler suspend the task).  This is caused
	// by a bug in the 386 kernel which needs to be fixed (i.e. PT_KILL
 	// allows some threads to return to user mode).
 	//
 	clear_trace_bit(current_thread);
#endif
	if (attach_flag) {
	    attach_flag = 0;
	    kill (inferior_pid, SIGHUP);
	    if (mach_call_no_error(task_resume(inferior_task),
		                   "task_resume", "kill_inferior_fast"))
		return;
	} else {
	    if (!stopped_in_ptrace) {
		kill (inferior_pid, SIGHUP);
		if (mach_call_no_error(task_resume(inferior_task),
		                       "task_resume", "kill_inferior_fast"))
		    return;
		if (!mach_oh_wait(&w)) {
		    ptrace(PT_KILL, inferior_pid, 0, 0);
		    mach_oh_wait(&w);
		}
	    } else {
		ptrace(PT_KILL, inferior_pid, 0, 0);
		mach_oh_wait(&w);
	    }
	}
    }
}

void
kill_inferior ()
{
    kill_inferior_fast ();
    NeXT_mourn_inferior ();
}

void
child_resume (step, signal)
     int step;
     int signal;
{
    kern_return_t ret;
    areStepping = step;
    errno = 0;
    [regionManager invalidate];
#ifdef IMPRECISE_EXECEPTIONS  
    if (breakPC) {
	[regionManager getDataAt: (void *)breakPC
			     for: sizeof(breakpoint_insn)
			    into: (void *)shadow];
	[regionManager putDataAt: (void *)breakPC
			     for: sizeof(breakpoint_insn)
			    from: (void *)breakpoint_insn];
        prepare_threads_before_run(0);
    } else
	prepare_threads_before_run(step);
#else /* IMPRECISE_EXECEPTIONS */
    prepare_threads_before_run(step);
#endif /* IMPRECISE_EXECEPTIONS */
    if (stopped_in_ptrace) {
	call_ptrace (step ? PTRACE_SINGLESTEP : PTRACE_CONT,
			inferior_pid, 1, signal);
	if (errno)
	    perror_with_name ("ptrace");
	else {
	    voyeurResume();
	    tell_client_program_is_running();
	}
    } else {
   if ((ret = task_resume (inferior_task)) != KERN_SUCCESS) {
	    fprintf(stderr, "Error resuming inferior.\n");
	    mach_error("task_resume", ret);
	    error("resume: task_resume");
	}
	voyeurResume();
	tell_client_program_is_running();
    }   
}

/* Start debugging the process whose number is PID.  */
int
attach (pid)
     int pid;
{
  errno = 0;
  attach_flag = 1;
  return pid;
}

/* Stop debugging the process whose number is PID
   and continue it with signal number SIGNAL.
   SIGNAL = 0 means just continue it.  */

void
detach (signal)
     int signal;
{
    unloadStatesFrom(0);
    removeLoadBreakpoint();
    if (kernel_attached) {
	mach_call(task_detach(inferior_task), "task_resume", "detach");
	kernel_attached = NO;
    } else
	mach_call(task_resume(inferior_task), "task_resume", "detach");
    attach_flag = 0;
}

/* Copy LEN bytes to or from inferior's memory starting at MEMADDR
   to debugger memory starting at MYADDR.   Copy to inferior if
   WRITE is nonzero.
  
   Returns the length copied, which is either the LEN argument or zero.
   This xfer function does not do partial moves, since child_ops
   doesn't allow memory operations to cross below us in the target stack
   anyway.  */

int
child_xfer_memory (CORE_ADDR memaddr, char *myaddr, int len, int write, struct target_ops *target)
{
  if (write)
    return [regionManager putDataAt: (void *)memaddr for: len from: myaddr];
  else {
    int ret = [regionManager getDataAt: (void *)memaddr for: len into: myaddr];
    if (!ret && read_memory_out_of_bounds_ok) {
//	printf ("Bad read attempt at %p for %d bytes.\n", memaddr, len);
	bzero(myaddr, len);
	ret = 1;
    }
    return ret;
  }
}

static void
badTaskFromRegionManager()
{
    NeXT_mourn_inferior();
    error("The program you were running has been terminated.");
} 

void 
NeXT_mourn_inferior ()
{
    static int inMourning = 0;
    
    if (inMourning) {
	// This means that we got an error while mourning.  Don't try it again.
	inMourning = 0;
    } else {
	inMourning = 1;
	tell_client_program_is_dead();
	unloadStatesFrom(0);
	inferior_task = TASK_NULL;
	regionManager = [regionManager free];
	clearPrintStream();
	child_mourn_inferior ();
	inMourning = 0;
	stopped_in_ptrace = 0;
    }
}

void
start_inferior_hook()
{
}

void
create_inferior_for_task(task_t task, BOOL isLazy)
{
    thread_array_t thread_list;
    unsigned int thread_count;
    extern int stop_print_frame;
    
    inferior_task = task;
    mach_call(
      port_allocate(task_self(), &inferior_exception_port),
      "port_allocate", "create_inferior_hook");
    mach_call(
      task_get_exception_port(inferior_task,  &inferior_old_exception_port),
      "task_get_special_port", "create_inferior_hook");
    mach_call(
      task_set_exception_port(inferior_task, inferior_exception_port),
      "task_set_special_port", "create_inferior_hook");
    mach_call_no_inferior(
      port_set_add(task_self(), gdb_ports, inferior_exception_port),
      "port_set_add", "create_inferior_hook");
    mach_call(
      task_threads(inferior_task, &thread_list, &thread_count),
      "task_threads", "create_inferior_hook");
    last_current_thread = current_thread = thread_list[0];
    mach_call(
      vm_deallocate(task_self(), (vm_address_t)thread_list, 
                    (vm_size_t)(thread_count * sizeof(int))),
      "vm_deallocate", "create_inferior_hook");
    regionManager = [(lazyRead || isLazy)
                     ? [LazyRegionManager class]
		     : [RegionManager class] 
			newTask: inferior_task readInRegions: NO];
    [regionManager setTaskGoneCallBack: &badTaskFromRegionManager];
    if (kernel_attached)
	[regionManager combineRegions: NO];
    current_target->to_wait = &mach_oh_wait;
    current_target->to_load = NULL;
    current_target->to_mourn_inferior = &NeXT_mourn_inferior;
//    insertLoadBreakpoint();
    clearThreadState();
}

task_t findPidServer(char *host)
{
    task_t pidServer = TASK_NULL;
    
    if (netname_look_up(name_server_port, host, "PIDSERVER", &pidServer)
	!= KERN_SUCCESS) {
	if (vfork()) {
	    int tries, found;
	    
	    for (tries = 0, found = NO; (tries < 20) && !found; tries++) {
		usleep(500000);
		found = (netname_look_up(name_server_port, host, "PIDSERVER", 
		                         &pidServer)
			 == KERN_SUCCESS);
	    }
	} else
	    execl("/usr/ucb/rsh", "/usr/ucb/rsh", host, "-n", 
	          "/usr/lib/pidServer", NULL);
    }
    return pidServer;
}

static void setToZero(int *pInt)
{
    *pInt = 0;
}

void
create_inferior_hook(char *args, int pid)
{
    task_t infTask;
    char *space, *host, *colon;
    struct cleanup *oldChain;
    
    oldChain = make_cleanup(setToZero, &inferior_pid);
    if (pid == -1) {
	colon = strchr(args, ':');
	if (colon) {
	    task_t pidServer;
	    *colon = '\0';
	    host = args;
	    pidServer = findPidServer(host);
	    if (!pidServer)
		error("Problem finding pid server on %s.", host);
	    else {
		int remotePid = atoi(colon + 1);
		
		if (task_by_unix_pid(pidServer, remotePid, &infTask)
		    == KERN_SUCCESS)
		    lazyRead = 1;
		else
		    error("Problem finding task for pid %d on %s",
			    remotePid, host);
	    }
	} else {
	    space = strchr(args, ' ');
	    if (space) {
		for (host = space; *host && (*host == ' '); host++);
		*space = '\0';
	    } else 
		host = "";
	    if (netname_look_up(name_server_port, host, args, &infTask)
		!= KERN_SUCCESS)
		    error("Problem finding task named %s on %s",
			    args,
			    *host ? host : "localhost");
	}
    } else {
	if (task_by_unix_pid(task_self(), pid, &infTask) != KERN_SUCCESS)
	    error("Problem attaching to pid %d.", pid);
    }

    kernel_attached = NO;
    create_inferior_for_task(infTask, NO);
    
    discard_cleanups(oldChain);
    
    if (attach_flag) {
	prepare_threads_after_stop();
	getLoadedSymbols();
	insertLoadBreakpoint();
	flush_cached_frames();
        registers_changed ();
	task_suspend(inferior_task);
	stop_pc = read_pc();
	step_start_function = find_pc_function(stop_pc);
	stop_print_frame = 1;
    } else {
	mutex_lock(childrenMutex);
	childrenExist = YES;
	condition_signal(childrenCondition);
	mutex_unlock(childrenMutex);
    }
}

static char *remoteHost = NULL;

int isTask(task_t task)
{
    struct task_basic_info taskInfo;
    unsigned int taskInfoCount;
    taskInfoCount = TASK_BASIC_INFO_COUNT;
    return ((task_info(task,
                       TASK_BASIC_INFO,
		       (task_info_t)&taskInfo,
		       &taskInfoCount) == KERN_SUCCESS) ? YES : NO);
}

void
suspendSignalThread()
{
    if (signalThread) {
	if (!signalSuspendCount)
	    thread_suspend(cthread_thread(signalThread));
	signalSuspendCount++;
    }
}

void
resumeSignalThread()
{
    if (signalThread) {
	if (signalSuspendCount)
	    thread_resume(cthread_thread(signalThread));
	signalSuspendCount--;
    }
}

void 
_initialize_next_inferior()
{
    mach_call(
      port_set_allocate(task_self(), &gdb_ports),
      "port_set_allocate", "_initialize_next_inferior");       
    mach_call(
      port_allocate(task_self(), &gdb_notify_port),
      "port_allocate", "_initialize_next_inferior");
    mach_call(
      task_set_notify_port(task_self(), gdb_notify_port),
      "task_set_notify_port",  "_initialize_next_inferior");
    mach_call(
      port_set_add(task_self(), gdb_ports, gdb_notify_port),
      "port_set_add", "_initialize_next_inferior");
    mach_call(
      port_allocate(task_self(), &signalPort),
      "port_allocate", "_initialize_next_inferior");
    mach_call(
      port_set_backlog(task_self(), signalPort, PORT_BACKLOG_MAX),
      "port_set_backlog", "_initialize_next_inferior");
    mach_call(
      port_set_add(task_self(), gdb_ports, signalPort),
      "port_set_add", "_initialize_next_inferior");
    childrenCondition = condition_alloc();
    childrenMutex = mutex_alloc();
    signalThread = cthread_fork((cthread_fn_t)&waitForSignal, NULL);
    cthread_set_name(signalThread, "signal thread");
    cthread_detach(signalThread);
#if 0    
    add_show_from_set(
      add_set_cmd("host", class_run, var_string_noescape, 
                (char *)&remoteHost,
                "Set name of host to run program on.\n"
		"Currently only works for mach tasks.",
		&setlist),
      &showlist);
#endif 0      	
    add_show_from_set(
      add_set_cmd("lazy-read", class_run, var_boolean, 
                (char *)&lazyRead,
                "Set if inferior's memory is read lazily.",
		&setlist),
      &showlist);		
}

void save_inferior_status_hook(struct inferior_status *inf_status)
{
    inf_status->extra_data[0] = current_thread;
    inf_status->extra_data[1] = step_thread;
}

void restore_inferior_status_hook(struct inferior_status *inf_status)
{
    current_thread = inf_status->extra_data[0];
    step_thread = inf_status->extra_data[1];
}
